"""Uploads RxNorm files.

This module provides methods to extract, transform and load relationships
defined by the RxNorm dataset.

The download files for RxNorm are provided at https://www.nlm.nih.gov/research/umls/rxnorm/.
"""
import os

from urllib.parse import urlparse

import petl as etl

from termlink.commands import SubCommand
from termlink.models import Coding, Relationship
from termlink.services import RelationshipService

_RXNCONSO_FIELDS = ["RXCUI", "LAT", "TS", "LUI", "STT", "SUI", "ISPREF", "RXAUI",
                    "SAUI", "SCUI", "SDUI", "SAB", "TTY", "CODE", "STR", "SRL", "SUPPRESS", "CVF", ]
_RXNREL_FIELDS = ["RXCUI1", "RXAUI1", "STYPE1", "REL", "RXCUI2", "RXAUI2",
                  "STYPE2", "RELA", "RUI", "SRUI", "SAB", "SL", "DIR", "RG", "SUPPRESS", "CVF", ]


def _to_json(rec):
    """
    Convert record in table to Relationship as a JSON object

    Record is expected to have the following fields: [ source.CODE, source.STR, 
    target.CODE, target.STR]

    Args:
        rec: A table record

    Returns:
        A new record containing a single field, which is the JSON object
    """

    source = Coding(
        system="http://www.nlm.nih.gov/research/umls/rxnorm",
        code=rec['source.CODE'],
        display=rec['source.STR']
    )

    target = Coding(
        system="http://www.nlm.nih.gov/research/umls/rxnorm",
        code=rec['target.CODE'],
        display=rec['target.STR']
    )

    relationship = Relationship('subsumes', source, target)

    return [relationship.to_json()]


class Command(SubCommand):
    """
    A command executor for RxNorm operations
    """

    @staticmethod
    def execute(args):
        """
        Prints a JSON array of `Relationship` objects to stdout

        Args:
            args: `argparse` parsed arguments
        """
        uri = urlparse(args.uri)
        service = Service(uri)
        table = service.get_relationships()
        etl.io.totext(table, encoding='utf8', template='{relationship}\n')


class Service(RelationshipService):
    """Converts the RxNorm database"""

    def __init__(self, uri):
        """
        Bootstraps a service

        Args:
            uri: URI to root location of .rrf files
        """

        if uri.scheme != 'file':
            raise ValueError("'uri.scheme' %s not supported" % uri.scheme)

        self.uri = uri

    def get_relationships(self):
        """
        Parses a list of `Relationship` objects.
        """
        path = os.path.join(self.uri.path, 'RXNCONSO.RRF')
        rxnconso = etl \
            .fromcsv(path, delimiter='|') \
            .setheader(_RXNCONSO_FIELDS) \
            .select(lambda rec: rec['SAB'] == 'RXNORM') \
            .cut('RXCUI', 'CODE', 'STR')

        source = rxnconso.prefixheader('source.')
        target = rxnconso.prefixheader('target.')

        path = os.path.join(self.uri.path, 'RXNREL.RRF')
        rxnrel = etl \
            .fromcsv(path, delimiter='|') \
            .setheader(_RXNREL_FIELDS) \
            .select(lambda rec: rec['SAB'] == 'RXNORM') \
            .select(lambda rec: rec['STYPE1'] == 'CUI') \
            .select(lambda rec: rec['STYPE2'] == 'CUI') \
            .select(lambda rec: rec['REL'] == 'RB') \
            .cut('RXCUI1', 'RXCUI2')

        return rxnrel \
            .join(source, lkey='RXCUI1', rkey='source.RXCUI') \
            .join(target, lkey='RXCUI2', rkey='target.RXCUI') \
            .rowmap(_to_json, ['source.CODE', 'source.STR', 'target.CODE', 'target.STR']) \
            .setheader(['relationship'])



"""A utility for iterating in batches.

This module provides utility methods for iterating over iterables in batches.
"""

import time

from termlink.configuration import Config

_configuration = Config()
_logger = _configuration.logger

_DEFAULT_RATE_LIMIT = 10
_DEFAULT_RATE_LIMIT_PERIOD = 1

_rate_limit = int(_configuration.get_property('API_RATE_LIMIT', _DEFAULT_RATE_LIMIT))
_rate_limit_period = int(_configuration.get_property('API_RATE_LIMIT_PERIOD', _DEFAULT_RATE_LIMIT_PERIOD))


def batch(iterable, n=1, sleep=_rate_limit_period):
    """
    Traverses over an iterable in batches of size n sleeping :sleep: seconds
    between batches.

    Args:
        iterable:   An iterable
        n:          Size of each batch
        sleep:      Number of seconds to sleep between batches

    Returns:
        For each batch, yields a batch of size n, len(iterable) % n times
    """

    limit = n / (_rate_limit)
    sleep = sleep if 0 < sleep < limit else limit
    _logger.info("sleep set to %f seconds", sleep)

    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx : min(ndx + n, l)]
        _logger.info("Processed %s of %s", min(ndx + n, l), l)
        time.sleep(sleep)



"""Services

A collection of interfaces built for extending ontology transformations.
"""

from abc import ABC, abstractmethod
from typing import List

from termlink.models import Relationship


class RelationshipService(ABC):
    """
    An interface for actions related to the `Relationship` model
    """

    @abstractmethod
    def get_relationships(self) -> List[Relationship]:
        """
        Gets a `List` of `Relationship` objects

        Returns:
            A `List` of `Relationship` objects
        """
        pass



"""The main program.

This module is the command line interface for running "termlink."
"""
import argparse

from termlink.configuration import Config

from termlink.rxnorm import Command as RxNormCommand

configuration = Config()
logger = configuration.logger

parser = argparse.ArgumentParser(
    description="""
    An ontology conversion toolkit for the Precision Health Cloud.
    """
)

subparsers = parser.add_subparsers(
    title="Commands",
    metavar=""
)

parser_rxnorm = subparsers.add_parser(
    "rxnorm",
    help="RxNorm provides normalized names for clinical drugs.",
    description="""
    RxNorm provides normalized names for clinical drugs and links its names to
    many of the drug vocabularies commonly used in pharmacy management and drug
    interaction software, including those of First Databank, Micromedex, 
    Gold Standard Drug Database, and Multum. By providing links between these 
    vocabularies, RxNorm can mediate messages between systems not using the 
    same software and vocabulary. [1] 
    """,
    epilog="""
    [1] RxNorm. Retrieved April 22, 2019, from https://www.nlm.nih.gov/research/umls/rxnorm/
    """
)
parser_rxnorm.add_argument(
    "uri", metavar="URI", help="resource identifier for files")
parser_rxnorm.set_defaults(execute=RxNormCommand.execute)

args = parser.parse_args()

if hasattr(args, 'execute'):
    args.execute(args)
else:
    parser.print_help()



"""Commands

This module is built for constructing command patterns.
"""

from abc import ABC, abstractmethod

class SubCommand(ABC):
    """
    An interface for create sub-commands
    """

    @staticmethod
    @abstractmethod
    def execute(args):
        """
        Executes based on the provided arguments

        Args:
            args: `argparse` parsed arguments
        """



"""Verifies the 'configuration.py' module."""
from nose.tools import ok_

from termlink.configuration import Config

configuration = Config()


def test_configuration_is_loaded():
    """Checks that the configuration is loaded"""
    ok_(configuration)


def test_logger_is_configured():
    """Checks that the application logger has been configured"""
    ok_(configuration.logger)



"""Provides client libraries for communicating with the LifeOmic API"""

from urllib.parse import urljoin, urlparse

from ratelimit import limits, RateLimitException
from backoff import on_exception, expo

from requests import Session

from termlink.adapter import DEFAULT as DEFAULT_ADAPTER
from termlink.configuration import Config


_DEFAULT_RATE_LIMIT = 10
_DEFAULT_RATE_LIMIT_PERIOD = 1
_DEFAULT_RATE_LIMIT_MAX_RETRIES = 3

_configuration = Config()

_rate_limit = int(_configuration.get_property('API_RATE_LIMIT', _DEFAULT_RATE_LIMIT))
_rate_limit_period = int(_configuration.get_property('API_RATE_LIMIT_PERIOD', _DEFAULT_RATE_LIMIT_PERIOD))
_rate_limit_max_retries = int(_configuration.get_property('API_RATE_LIMIT_MAX_RETIRES', _DEFAULT_RATE_LIMIT_MAX_RETRIES))


class Client:
    """An HTTP client"""

    def __init__(
            self,
            url,
            adapter=DEFAULT_ADAPTER):
        """
        Creates a new API client.

        Args:
            adapter (HTTPAdapter):  an HTTPAdapter
        """

        if url is None:
            raise TypeError("'url' is required")

        session = Session()

        parsed = urlparse(url)
        prefix = parsed[0] + "://"
        session.mount(prefix, adapter)

        self.session = session
        self.url = url

    @on_exception(expo, RateLimitException, max_tries=_rate_limit_max_retries)
    @limits(calls=_rate_limit, period=_rate_limit_period)
    def request(self, method, path=None, data=None):
        """
        A facade around :func:`requests.request` that provides rate limiting.

        Args:
            method: request method
            path:   URL path
            data:   JSON  data

        Returns:
            A :class:`Response <Response>` object
        """
        url = urljoin(self.url, path) if path else self.url
        return self.session.request(method, url, json=data)



"""
Utility method for loading properties form the configuration file and environment variables.
"""

import configparser
import logging
import os

# Relative location of configuration file
_PROGRAM_CONFIG = '../config.ini'
_LOGGER_FILENAME = 'stderr.log'

_environment = os.getenv("ENV", 'DEFAULT').upper()


class Config:
    """A configuration singleton"""

    class __Config:
        """A configuration supplier.

        An object for reading configuration properties. Properties are injected with
        the following priority. Properties with higher priority overwrite those
        with lower priority.

        1. Program defaults
        2. The 'DEFAULT' section in 'config.ini'
        3. The environment in the 'config.ini' file set by the 'ENV' environment variable
        4. Runtime system environment variables
        """

        def __init__(self):

            # Create an application logger
            logger = logging.getLogger("termlink")
            logging.basicConfig(filename=_LOGGER_FILENAME)

            # Set the logging level
            if _environment == "TEST":
                logger.setLevel(logging.DEBUG)
            else:
                logger.setLevel(logging.INFO)

            logger.info("The logging level has been set to %s", logger.level)

            parser = configparser.ConfigParser(os.environ, strict=False)

            # Get the absolute path based on the execution directory
            root = os.path.abspath(os.path.dirname(__file__))
            config = os.path.join(root, _PROGRAM_CONFIG)

            configs = [config]
            parser.read(configs)

            self.logger = logger
            self.parser = parser

    instance = None

    def __init__(self):
        if not Config.instance:
            Config.instance = Config.__Config()

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def get_property(self, property_name, default=None):
        """Get a property value from the configuration.

        The os environment variables are first checked. If the property
        does not exists there it is read from the configuration file.

        Args:
            property_name (str):    A name, or key, of a property

        Returns:
            The property value associated with the property_name
        """
        if property_name in os.environ:
            return os.environ[property_name]

        return self.parser.get(_environment, property_name, fallback=default)



"""Validates the 'adapter.py' module"""

from nose.tools import ok_

from urllib3.util.retry import Retry

from termlink.adapter import build

def test_build():
    """Checks that build works be default"""
    adapter = build()
    ok_(adapter)

def test_build_with_custom_retry_policy():
    """Checks that an adapter can be built with a custom retry policy"""
    retry_policy = Retry()
    adapter = build(retry_policy=retry_policy)
    ok_(adapter)



"""Verifies the 'models.py' module"""

import json

from nose.tools import eq_

from termlink.models import Coding, Relationship


def test_convert_coding_to_json():
    """Checks converting `Coding` to JSON"""

    system = 'system'
    version = 'version'
    code = 'code'
    display = 'display'

    coding = Coding(
        system=system,
        version=version,
        code=code,
        display=display
    )

    exp = {
        'type': 'coding',
        system: system,
        version: version,
        code: code,
        display: display
    }

    res = coding.to_json()

    eq_(json.dumps(exp), res)


def test_convert_empty_coding_to_json():
    """Checks converting `Coding` with no values to JSON"""

    coding = Coding()

    exp = {
        'type': 'coding',
        'system': None,
        'version': None,
        'code': None,
        'display': None
    }
    res = coding.to_json()

    eq_(json.dumps(exp), res)


def test_convert_partial_coding_to_json():
    """Checks converting `Coding` with some values to JSON"""

    system = 'system'

    coding = Coding(system=system)

    exp = {
        'type': 'coding',
        'system': system,
        'version': None,
        'code': None,
        'display': None
    }

    res = coding.to_json()

    eq_(json.dumps(exp), res)


def test_convert_relationship_to_json():
    """Checks converting `Relationship` to JSON"""

    equivalence = 'equivalence'
    system = 'system'
    version = 'version'
    code = 'code'
    display = 'display'

    source = Coding(
        system=system,
        version=version,
        code=code,
        display=display
    )

    target = Coding(
        system=system,
        version=version,
        code=code,
        display=display
    )

    relationship = Relationship(equivalence, source, target)

    exp = {
        'equivalence': equivalence,
        'source': {
            'type': 'coding',
            system: system,
            version: version,
            code: code,
            display: display
        },
        'target': {
            'type': 'coding',
            system: system,
            version: version,
            code: code,
            display: display
        }
    }

    res = relationship.to_json()

    eq_(json.dumps(exp), res)



"""A module with configured adapaters"""

from requests.adapters import HTTPAdapter

_DEFAULT_POOL_SIZE = 1

DEFAULT = HTTPAdapter()


def build(
        min_pool_size=_DEFAULT_POOL_SIZE,
        max_pool_size=_DEFAULT_POOL_SIZE,
        is_blocking=False,
        retry_policy=0):
    """A helper method to build an adapter

    Args:
        min_pool_size (int):    minimum connection pool size
        max_pool_size (int):    maximum connection pool size
        blocking (bool):        if the connection pool is blocking
        retry_policy (Retry):   a retry policy

    Returns:
        An :class:`HTTPAdapter` configured a specified
    """
    return HTTPAdapter(
        pool_connections=min_pool_size,
        pool_maxsize=max_pool_size,
        pool_block=is_blocking,
        max_retries=retry_policy
    )



"""Verifies the 'rxnorm.py' module"""

from urllib.parse import urlparse

from nose.tools import ok_, raises

from termlink.rxnorm import Service


def test_service_uri_can_be_file():
    """Checks that a uri.scheme of 'file' is ok"""
    uri = urlparse("file://")
    ok_(Service(uri=uri))


@raises(ValueError)
def test_service_uri_requires_scheme_file():
    """Checks that a uri.scheme of 'foobar' throws a ValueError"""
    uri = urlparse("foobar://")
    Service(uri=uri)



"""Validates the 'client.py" module"""

from nose.tools import ok_, eq_

from requests_mock import Adapter

from termlink.client import Client


def test_client_construction():
    """Checks that the client can be instantiated"""

    url = 'url'

    client = Client(url)

    ok_(client)


def test_mocking_client_adapter():
    """Checks that mocking the client adapter is possible"""

    url = 'http://mock.com'

    adapter = Adapter()
    client = Client(url=url, adapter=adapter)

    exp = 'success'
    adapter.register_uri('GET', url, text=exp)
    res = client.request('get')

    eq_(200, res.status_code)
    eq_(exp, res.text)


def test_rate_limit():
    """Checks that the rate limit is working"""

    url = 'http://mock.com'

    adapter = Adapter()
    client = Client(url=url, adapter=adapter)

    exp = 'success'
    adapter.register_uri('GET', url, text=exp)
    res = client.request('get')

    eq_(200, res.status_code)
    eq_(exp, res.text)



#!/usr/bin/env python

import json

from setuptools import setup, find_packages

_PACKAGE_FILE = './package.json'
_README_FILE = './README.md'
_REQUIREMENTS_FILE = './requirements.txt'

with open(_PACKAGE_FILE, 'r') as file:
    package = json.load(file)

with open(_README_FILE, "r") as file:
    long_description = file.read()

with open(_REQUIREMENTS_FILE, 'r') as file:
    requirements = file.read().splitlines()

print(long_description)

setup(
    name=package['name'],
    version=package['version'],
    description=package['description'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=package['url'],
    license=package['license'],
    author='Taylor Steinberg',
    author_email='taylor.steinberg@lifeomic.com',
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Intended Audience :: Healthcare Industry',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Topic :: Utilities'
    ],
)



"""Model representations of API interfaces"""

from dataclasses import dataclass
from dataclasses_json import DataClassJsonMixin

from termlink.configuration import Config

_configuration = Config()
_logger = _configuration.logger


@dataclass(frozen=True)
class Coding(DataClassJsonMixin):
    """
    A 'Coding' object as defined by the API.

    Attributes:
        system (str):   Identity of the terminology system
        version (str):  Version of the system
        code (str):     Symbol in syntax defined by the system
        display (str):  Representation defined by the system
    """
    type: str = 'coding'
    system: str = None
    version: str = None
    code: str = None
    display: str = None


@dataclass
class Relationship(DataClassJsonMixin):
    """
    A 'Relationship' object as defined by the API.

    Attributes:
        equivalence (str):  The degree of equivalence between concepts.
        source object:       The source concept.
        target object:       The target concept.
    """
    equivalence: str
    source: Coding
    target: Coding
